﻿using System;
using System.Collections.Generic;
using System.Text;

using CommonMP.HYSSOP.CoreImpl;
using CommonMP.HYSSOP.CoreImpl.HSData;
using CommonMP.HYSSOP.CoreImpl.HSTools;

using CommonMP.HYMCO.Interface;
using CommonMP.HYMCO.Interface.Model;
using CommonMP.HYMCO.Interface.Data;
using CommonMP.HYMCO.CoreImpl.Data;
using CommonMP.HYMCO.CoreImpl.Tool;



namespace KyotoUnivEngHywr
{
	/// <summary><para>class outline:</para>
	/// <para>KyotoUnivRiverKinematicWaveModel 河道流追跡モデルクラス</para>
	/// </summary>
    public class RoutingModelWithKinematicWaveModel : McForecastModelBase
	{
		/// <summary> 演算データ（キャスト用） </summary>
        RoutingModelWithKinematicWaveInfo mInf = null;

		/// <summary><para>method outline:</para>
		/// <para>入力側の接続情報チェック</para>
		/// </summary>
		/// <example><para>usage:</para>
		/// <para>bool bRtn = ReceiveConnectionCheck(ref csErrorInf)</para>
		/// </example>
		/// <param name="csErrorInf">エラー出力</param>
		/// <param name="lInputDataNum">入力情報数</param>
		/// <param name="csInputCellData">演算に必要な入力情報配列</param>
		/// <returns>=true:正常、=false:異常</returns>
		/// <exception cref="">無し</exception>
		/// <remarks><para>remarks:</para>
		/// <para>受信するデータが自モデルが期待している情報か否かをチェックする</para>
		/// </remarks>
		protected override bool ReceiveConnectionCheck(ref McStructErrorInfo csErrorInf, long lInputDataNum, McReceiveCellDataIF[] csInputCellData)
		{
			bool bRtn = true;

			if (lInputDataNum == 0)
			{
				// 入力が無い場合エラー
                csErrorInf.AddCheckErrorData(this.GetID(), RoutingModelWithKinematicWaveDefine.HYDROLOGY_RIVER_KINEMATIC_WAVE_KIND, "No receive data.");
				bRtn = false;
			}
			else if (lInputDataNum > 0)
			{
				mInf.mNumUpInflow = 0;
				mInf.mNumLateral = 0;

				for (long lP = 0; lP < lInputDataNum; lP++)
				{   // 入力データ数分

                    if (RoutingModelWithKinematicWaveDefine.IN_PATTERN_TOP.Equals(csInputCellData[lP].GetReceivePatternID()))
					{
						mInf.mNumUpInflow++;
						
					}
                    else if (RoutingModelWithKinematicWaveDefine.IN_PATTERN_LATERAL.Equals(csInputCellData[lP].GetReceivePatternID()))
					{
						mInf.mNumLateral++;
					}
					else
					{
						// 予期しない接続
                        csErrorInf.AddCheckErrorData(this.GetID(), RoutingModelWithKinematicWaveDefine.HYDROLOGY_RIVER_KINEMATIC_WAVE_KIND,
							"Unexpected receive data type. (Received from " + csInputCellData[lP].GetUpperElementID().ToString() + ")");
						bRtn = false;
					}
				}
			}
			return bRtn;
		}
		/// <summary><para>method outline:</para>
		/// <para>出力側の接続情報チェック</para>
		/// </summary>
		/// <example><para>usage:</para>
		/// <para>bool bRtn = SendConnectionCheck(ref csErrorInf)</para>
		/// </example>
		/// <param name="csErrorInf">エラー出力</param>
		/// <param name="lOutputDataNum">出力情報数</param>
		/// <param name="csOutputCellData">出力情報配列</param>
		/// <returns>=true:正常、=false:異常</returns>
		/// <exception cref="">無し</exception>
		/// <remarks><para>remarks:</para>
		/// <para>送信端子に設定されている伝送データが自モデルが期待している情報か否かをチェックする</para>
		/// </remarks>
		protected override bool SendConnectionCheck(ref McStructErrorInfo csErrorInf, long lOutputDataNum, McSendCellDataIF[] csOutputCellData)
		{
			bool bRtn = true;

			if (lOutputDataNum == 0)
			{
				// 警告（エラーではない）
                csErrorInf.AddCheckWarningData(this.GetID(), RoutingModelWithKinematicWaveDefine.HYDROLOGY_RIVER_KINEMATIC_WAVE_KIND, "No send port.");
			}
			else if (lOutputDataNum > 0)
			{
				for (long lP = 0; lP < lOutputDataNum; lP++)
				{   // 入力データ数分

                    if (RoutingModelWithKinematicWaveDefine.OUT_PATTERN_BOTTOM.Equals(csOutputCellData[lP].GetSendPatternID()) == true)
					{
						// 問題なし
					}
					else
					{
						// 予期しない接続
                        csErrorInf.AddCheckErrorData(this.GetID(), RoutingModelWithKinematicWaveDefine.HYDROLOGY_RIVER_KINEMATIC_WAVE_KIND,
							"Unexpected send data type. (Send To " + csOutputCellData[lP].GetLowerElementID().ToString() + ")");
						bRtn = false;
					}
				}
			}

			return bRtn;
		}

		/// <summary><para>method outline:</para>
		/// <para>モデルを初期化する</para>
		/// </summary>
		/// <example><para>usage:</para>
		/// <para>Initialize(csInitialData)</para>
		/// </example>
		/// <param name="csInitialData">初期化設定情報</param>
		/// <param name="lInputDataNum">入力情報数</param>
		/// <param name="csInputCellData">演算に必要な入力情報配列</param>
		/// <returns>無し</returns>
		/// <exception cref="">無し</exception>
		/// <remarks><para>remarks:</para>
		/// <para>無し</para>
		/// </remarks>
		protected override bool Initialize(ref McPropertyInfoRoot csInitialData, long lInputDataNum, ref McReceiveCellDataIF[] csInputCellData)
		{
			McLog.DebugOut(m_csSimTime, GetID(), GetType().ToString(), "Initialize", "<<Init>>");

			// 引数で与えられたデータを　キャストしています。
			McInitialInfo csInDt = csInitialData as McInitialInfo;
			
			if (csInDt != null)
			{
				double isTotal = 0.0;
				long mDiv = mInf.mDiv;

				for (int i = 0; i <= mInf.mDiv; i++)
				{
					String mkname = "mQc-" + i.ToString();
					csInDt.GetInfo(mkname, ref mInf.mQc[i]);
					isTotal += mInf.mQc[i];
				}

				// 初期状態量が設定されていないとみられる場合
				if (isTotal <= 0.0001)
				{
					HySCellData csCell = null;
					mInf.mQlc = 0.0;
					mInf.mQc[0] = 0.0;
					for (long lP = 0; lP < lInputDataNum; lP++)
					{   // 入力データ数分
                        if (RoutingModelWithKinematicWaveDefine.IN_PATTERN_TOP.Equals(csInputCellData[lP].GetReceivePatternID()) == true)
						{
							// 上流端からの流入量 (m3/s)
							csCell = csInputCellData[lP].GetInterpolatedCell(0);
							mInf.mQc[0] += csInputCellData[0].Data(csCell, 0);
						}
                        else if (RoutingModelWithKinematicWaveDefine.IN_PATTERN_LATERAL.Equals(csInputCellData[lP].GetReceivePatternID()))
						{
							// 側方流入量 (m2/s)
							csCell = csInputCellData[lP].GetInterpolatedCell(0);
							mInf.mQlc += csInputCellData[0].Data(csCell, 0);
						}
						else
						{
						}
					}

					mInf.mQc[mDiv] = mInf.mQc[0] + mInf.mQlc;
					mInf.mQlc /= mInf.mL;

					for (int i = 0; i <= mDiv; i++)
					{
						mInf.mQc[i] = mInf.mQc[0] + (mInf.mQc[mDiv] - mInf.mQc[0]) * i / mDiv;
					}
				}

				for (int i = 0; i <= mDiv; i++)
				{
					mInf.mAc[i] = Math.Pow(mInf.mQc[i] / mInf.mAlpha[i] , 1.0 / mInf.mM[i]);
				}
			}
			
			return true;
		}

		// =======================
		// 演算実行処理関連メソッド
		// =======================

		/// <summary><para>method outline:</para>
		/// <para>モデルを初期化する</para>
		/// </summary>
		/// <example><para>usage:</para>
		/// <para>OnlineInitialize(csInitialData)</para>
		/// </example>
		/// <param name="csInitialData">初期化設定情報</param>
		/// <param name="lInputDataNum">入力情報数</param>
		/// <param name="csInputCellData">演算に必要な入力情報配列</param>
		/// <returns>bool</returns>
		/// <exception cref="">無し</exception>
		/// <remarks><para>remarks:</para>
		/// <para>無し</para>
		/// </remarks>
		protected override bool OnlineInitialize(ref McPropertyInfoRoot csInitialData, long lInputDataNum, ref McReceiveCellDataIF[] csInputCellData)
		{
			// Do Nothing
			return true;
		}

		/// <summary><para>method outline:</para>
		/// <para>モデル演算</para>
		/// </summary>
		/// <example><para>usage:</para>
		/// <para>long lRtn = Calculate(ref csInputCellData)</para>
		/// </example>
		/// <param name="lInputDataNum">入力情報数</param>
		/// <param name="csInputCellData">演算に必要な入力情報配列</param>
		/// <returns>=0:正常 -1:異常</returns>
		/// <exception cref="">無し</exception>
		/// <remarks><para>remarks:</para>
		/// <para>無し</para>
		/// </remarks>
		protected override long Calculate(long lInputDataNum, ref McReceiveCellDataIF[] csInputCellData)
		{
			McLog.DebugOut(m_csSimTime, GetID(), GetType().ToString(), "Calculate", "<<Start>>");
			HySCellData csCell = null;
			mInf.mQln = 0.0;
			mInf.mQn[0] = 0.0;
			for (long lP = 0; lP < lInputDataNum; lP++)
			{
				csInputCellData[lP].SetCurrentTime(m_csSimTime + m_csDltTime);
				csCell = csInputCellData[lP].GetInterpolatedCell(0);

				// 入力データ数分			
                if (RoutingModelWithKinematicWaveDefine.IN_PATTERN_TOP.Equals(csInputCellData[lP].GetReceivePatternID()) == true)
				{
					// 上流端からの流入量 (m3/s)
					mInf.mQn[0] += csInputCellData[lP].Data(csCell, 0); 
				}
				else
				{
					// 側方流入量 (m2/s)
					mInf.mQln += csInputCellData[lP].Data(csCell, 0); 
				}
			}


			double dt = mInf.mDt; // time_step;

			mInf.mQln /= mInf.mL;
			double wQl = 0.5 * (mInf.mQlc + mInf.mQln);

			mInf.mAn[0] = Math.Pow(mInf.mQn[0] / mInf.mAlpha[0], 1.0 / mInf.mM[0]);

			for (int i = 1; i <= mInf.mDiv; i++)
			{
				double wCa = mInf.mAlpha[i - 1] * mInf.mM[i - 1] * Math.Pow(mInf.mAc[i - 1], mInf.mM[i - 1] - 1.0);
				
				double wQ = Math.Max(mInf.mQc[i], 0.5 * (mInf.mQc[i - 1] + mInf.mQn[i - 1]) + wQl * mInf.mDx);
				double wCb =
						Math.Pow(mInf.mAlpha[i], 1.0 / mInf.mM[i]) * mInf.mM[i] * Math.Pow(wQ, (mInf.mM[i] - 1.0) / mInf.mM[i]);
				
				double wC = 0.5 * (wCa + wCb);

				mInf.mQn[i] = (mInf.mQc[i] + mInf.mTheta * wC * (dt / mInf.mDx * mInf.mQn[i - 1] + dt * wQl) -
					(1.0 - mInf.mTheta) * wC * dt * ((mInf.mQc[i] - mInf.mQc[i - 1]) / mInf.mDx - wQl)) /
					(1.0 + mInf.mTheta * wC * dt / mInf.mDx);

				if (mInf.mQn[i] < 0)	mInf.mQn[i] = 0;
				
				mInf.mAn[i] = Math.Pow(mInf.mQn[i] / mInf.mAlpha[i], 1.0 / mInf.mM[i]);
			}

			// 状態量の更新
			for (int i = 0; i <= mInf.mDiv; i++)
			{
				mInf.mQc[i] = mInf.mQn[i];
				mInf.mAc[i] = mInf.mAn[i];
				mInf.mQlc = mInf.mQln;
			}

			
			return 0;
		}

		/// <summary><para>method outline:</para>
		/// <para>モデル演算結果を外部のエレメントに対して公開する</para>
		/// </summary>
		/// <example><para>usage:</para>
		/// <para>long lRtn = DataFusion( lOutputDataNum, ref csOutputCellData)</para>
		/// </example>
		/// <param name="lOutputDataNum">出力情報数</param>
		/// <param name="csOutputCellData">出力情報配列</param>
		/// <returns>=0:正常 -1:異常</returns>
		/// <exception cref="">無し</exception>
		/// <remarks><para>remarks:</para>
		/// <para>無し</para>
		/// </remarks>
		protected override long DataFusion(long lOutputDataNum, ref McSendCellDataIF[] csOutputCellData)
		{
			McLog.DebugOut(m_csSimTime, GetID(), GetType().ToString(), "DataFusion", "<<Start>>");
			
			HySCellData[] csSndCellData = null;
			// 最終セルだけの出力
			for (long lP = 0; lP < lOutputDataNum; lP++)
			{
				csSndCellData = csOutputCellData[lP].PrepareSendCellD1();
				csSndCellData[0].m_dData[0] = mInf.mQc[mInf.mDiv];
			}
			return 0;
		}

		// ====================
		// その他必要なメソッド
		// ====================

		/// <summary><para>method outline:</para>
		/// <para>プロパティ情報を設定する</para>
		/// </summary>
		/// <example><para>usage:</para>
		/// <para>SetProperty(csCellMdlPropertyInfo)</para>
		/// </example>
		/// <param name="csCellMdlPropertyInfo">セル型プロパティ情報</param>
		/// <returns>無し</returns>
		/// <exception cref="">無し</exception>
		/// <remarks><para>remarks:</para>
		/// <para>無し</para>
		/// </remarks>
		public override bool SetProperty(McCellModelPropertyIF csCellMdlPropertyInfo)
		{
			McLog.DebugOut(m_csSimTime, GetID(), GetType().ToString(), "SetProperty", "<<Init>>");
			// 使用しやすいようにキャストしておく
            mInf = (RoutingModelWithKinematicWaveInfo)m_csCalInfo;

			// プロパティ設定
			McCellModelPropertyInfo csPrptyInfo = csCellMdlPropertyInfo as McCellModelPropertyInfo;
			 
			if (csPrptyInfo != null)
			{
				// 演算ステップ時刻設定
				this.m_csDltTime = new HySTime(csPrptyInfo.GetStepTime());

				csPrptyInfo.GetInfo("mNumLateral", ref mInf.mNumLateral);
				csPrptyInfo.GetInfo("mNumUpInflow", ref mInf.mNumUpInflow);

				csPrptyInfo.GetInfo("mL", ref mInf.mL);
				csPrptyInfo.GetInfo("mB", ref mInf.mB);
				csPrptyInfo.GetInfo("mG", ref mInf.mG);
				csPrptyInfo.GetInfo("mMan", ref mInf.mMan);

				long iDiv = 0;
				csPrptyInfo.GetInfo("mDiv", ref iDiv);

				mInf.mDt = (long)(this.m_csDltTime.TotalSec());
				
				csPrptyInfo.GetInfo("mTheta", ref mInf.mTheta);

				double dSloop = 2000.0;
				csPrptyInfo.GetInfo("m_dSlope", ref dSloop);

				mInf.Initialize_mDiv(iDiv);
				for (int i = 0; i <= iDiv; i++)
				{
					String mkname = "mK-" + i.ToString();
					String mzname = "mZ-" + i.ToString();

					csPrptyInfo.GetInfo(mkname, ref mInf.mK[i]);
					csPrptyInfo.GetInfo(mzname, ref mInf.mZ[i]);
				}

				mInf.mDx = mInf.mL / mInf.mDiv;
				mInf.mG = Math.Atan(mInf.mG);
				mInf.mG = Math.Sin(mInf.mG); // sin の値に変換

				for (int i = 0; i <= mInf.mDiv; i++)
				{
					mInf.mAlpha[i] = 1.0 / mInf.mMan * Math.Pow(mInf.mK[i], 0.666667) * Math.Sqrt(mInf.mG);
					mInf.mM[i] = 1.0 + 0.666667 * mInf.mZ[i];
				}

			}
			return true;
		}
		/// <summary><para>method outline:</para>
		/// <para>演算実行中断中のプロパティ等情報設定</para>
		/// </summary>
		/// <example><para>usage:</para>
		/// <para>bool bRtn = SetOnlineProperty(csCellMdlPropertyInfo)</para>
		/// </example>
		/// <param name="csCellMdlPropertyInfo">プロパティ情報</param>
		/// <returns>無し</returns>
		/// <exception cref="">無し</exception>
		/// <remarks><para>remarks:</para>
		/// <para>演算中断中にプロパティ、初期値等を変更して動作させ場合等に使用する</para>
		/// </remarks>
		public override bool SetOnlineProperty(McCellModelPropertyIF csCellMdlPropertyInfo)
		{
			// プロパティ設定
			McCellModelPropertyInfo csPrptyInfo = csCellMdlPropertyInfo as McCellModelPropertyInfo;
			if (csPrptyInfo != null)
			{
				// To Do
				// モデルの係数等、必要な情報に対して、引数で与えられたプロパティ情報の内容を読み取って、設定していきます。
				// 但し、実行中に変更して困る情報は設定しないで下さい。
			}
			return true;
		}
	}
}
